import {
  ProtocolNotification,
  ProtocolRequest,
  ProtocolResponse,
} from './protocol';

import {
  PROTOCOL_NAME as RPCProtocol,
  RPCRequest,
  RPCResponse,
} from './protocol/rpc';
import injectable from '../context/decorator/injectable';

export type OnConnectFunc = () => void;
export type OnDisconnectFunc = () => void;
export type OnResponseFunc = (response: any) => void;

type ID_TYPE = number;

interface RequestQueue {
  resolve: (value?: any) => void;
  reject: (reason?: any) => void;
}

@injectable()
export default class WebSocketRPC {
  private url: string;
  private conn: WebSocket;
  private isReady: boolean;
  private requestID: ID_TYPE = 0;

  private onConnectListeners: OnConnectFunc[] = [];
  private onDisconnectListeners: OnDisconnectFunc[] = [];
  private onResponseListeners: Map<string, OnResponseFunc[]>;

  private requestQueue: Map<ID_TYPE, RequestQueue>;
  /**
   * name
   */
  public constructor(url: string) {
    this.isReady = false;
    this.url = url;
    this.onResponseListeners = new Map();
    this.requestQueue = new Map();

    this.initialize();
  }

  public Call(method: string, params: any[]): Promise<any> {


    return new Promise<any>((resolve, reject) => {
      const requestID = this.getRequestID();
      let request = new ProtocolRequest<ID_TYPE>(RPCProtocol, requestID);
      let req = new RPCRequest(method, params);
      request.setBody(req.toJSON());
      this.conn.send(request.toJSON());
      this.requestQueue.set(requestID, {resolve: resolve, reject: reject});
    });
  }


  public OnConnect(cb: OnConnectFunc): void {
    if (this.isReady) {
      cb();
    }
    this.onConnectListeners.push(cb);
  }

  public OnDisconnect(cb: OnDisconnectFunc): void {
    this.onDisconnectListeners.push(cb);
  }

  private initialize(): void {
    this.conn = new WebSocket(this.url);
    this.conn.onopen = (e: Event): any => {
      this.fireOnConnect();
      this.isReady = true;
      return null;
    };

    this.conn.onclose = (e: CloseEvent): any => {
      this.fireOnDisconnect();
      return null;
    };

    this.conn.onmessage = (e: MessageEvent): any => {
      let response: ProtocolResponse<ID_TYPE> = JSON.parse(e.data);
      let res: RPCResponse = JSON.parse(response.getBody());
      const requestID = response.getID();
      const error = response.getError();
      const result = res.getResult();

      if (this.requestQueue.has(requestID)) {
        let promise = this.requestQueue.get(requestID);
        this.requestQueue.delete(requestID);
        if (null != result) {
          promise.resolve(result);
        } else if (null != error) {
          promise.reject(error);
        } else {
          console.log('??????');
        }
      } else {
        // perhaps sever side send message
      }

      this.fireOnMessage(e);
    };

  }

  private fireOnConnect(): void {
    for (let i = 0; i < this.onConnectListeners.length; i++) {
        this.onConnectListeners[i]();
    }
  }

  private fireOnDisconnect(): void {
    for (let i = 0; i < this.onDisconnectListeners.length; i++) {
        this.onDisconnectListeners[i]();
    }
  }

  private fireOnMessage(e: MessageEvent): void {


    for (let i = 0; i < this.onDisconnectListeners.length; i++) {
        this.onDisconnectListeners[i]();
    }
  }

  private getRequestID(): number {
    return ++this.requestID;
  }

}
